from django.contrib.humanize.templatetags.humanize import intcomma
from django.urls import reverse
from django.utils.html import format_html
from django.utils.translation import gettext as _

from taxsystem.api.helpers import generate_button, get_info_button
from taxsystem.models.tax import Payments


def _payments_actions(corporation_id, payment: Payments, perms, request):
    actions = []
    if perms:
        amount = intcomma(payment.amount, use_l10n=True)
        if payment.is_pending or payment.is_needs_approval:
            actions.append(
                generate_button(
                    corporation_id=corporation_id,
                    template="taxsystem/partials/form/button.html",
                    queryset=payment,
                    settings={
                        "title": _("Approve Payment"),
                        "icon": "fas fa-check",
                        "color": "success",
                        "text": _("Approve Payment")
                        + f" {amount} ISK "
                        + _("from")
                        + f" {payment.account.user.username}",
                        "modal": "payments-approve",
                        "action": reverse(
                            viewname="taxsystem:approve_payment",
                            kwargs={
                                "corporation_id": corporation_id,
                                "payment_pk": payment.pk,
                            },
                        ),
                        "ajax": "action",
                    },
                    request=request,
                )
            )
            actions.append(
                generate_button(
                    corporation_id=corporation_id,
                    template="taxsystem/partials/form/button.html",
                    queryset=payment,
                    settings={
                        "title": _("Reject Payment"),
                        "icon": "fas fa-times",
                        "color": "danger",
                        "text": _("Reject Payment")
                        + f" {amount} ISK "
                        + _("from")
                        + f" {payment.account.user.username}",
                        "modal": "payments-reject",
                        "action": reverse(
                            viewname="taxsystem:reject_payment",
                            kwargs={
                                "corporation_id": corporation_id,
                                "payment_pk": payment.pk,
                            },
                        ),
                        "ajax": "action",
                    },
                    request=request,
                )
            )
        elif payment.is_approved or payment.is_rejected:
            actions.append(
                generate_button(
                    corporation_id=corporation_id,
                    template="taxsystem/partials/form/button.html",
                    queryset=payment,
                    settings={
                        "title": _("Undo Payment"),
                        "icon": "fas fa-undo",
                        "color": "danger",
                        "text": _("Undo Payment")
                        + f" {amount} ISK "
                        + _("from")
                        + f" {payment.account.user.username}",
                        "modal": "payments-undo",
                        "action": reverse(
                            viewname="taxsystem:undo_payment",
                            kwargs={
                                "corporation_id": corporation_id,
                                "payment_pk": payment.pk,
                            },
                        ),
                        "ajax": "action",
                    },
                    request=request,
                )
            )
    if payment.account.user == request.user or perms:
        actions.append(get_info_button(corporation_id, payment, request))

    actions_html = format_html("".join(actions))
    return format_html('<div class="d-flex justify-content-end">{}</div>', actions_html)



"""
Helper functions for static integrity calculations
"""

import logging

# Standard Library
import os
from pathlib import Path

# Third Party
from sri import Algorithm, calculate_integrity

# AA Fleet Pings
from taxsystem import __title__
from taxsystem.constants import AA_TAXSYSTEM_STATIC_DIR

logger = logging.getLogger(__name__)


def calculate_integrity_hash(relative_file_path: str) -> str:
    """
    Calculates the integrity hash for a given static file
    :param self:
    :type self:
    :param relative_file_path: The file path relative to the `aa-taxsystem/taxsystem/static/taxsystem` folder
    :type relative_file_path: str
    :return: The integrity hash
    :rtype: str
    """

    file_path = os.path.join(AA_TAXSYSTEM_STATIC_DIR, relative_file_path)
    integrity_hash = calculate_integrity(Path(file_path), Algorithm.SHA512)

    return integrity_hash



# Django
import logging

from django.db import models

logger = logging.getLogger(__name__)


class LogsQuerySet(models.QuerySet):
    pass


class LogsManagerBase(models.Manager):
    pass


LogsManager = LogsManagerBase.from_queryset(LogsQuerySet)



"""Tax Helpers"""

import logging

from django.db import transaction
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from taxsystem.models.filters import SmartGroup
from taxsystem.models.logs import PaymentHistory
from taxsystem.models.tax import (
    OwnerAudit,
    Payments,
    PaymentSystem,
)
from taxsystem.models.wallet import CorporationWalletJournalEntry

logger = logging.getLogger(__name__)


# pylint: disable=too-many-locals
def update_corporation_payments(corp_id):
    """Update corporation members"""
    audit_corp = OwnerAudit.objects.get(corporation__corporation_id=corp_id)

    logger.debug(
        "Updating payments for: %s",
        audit_corp.corporation.corporation_name,
    )

    accounts = PaymentSystem.objects.filter(corporation=audit_corp)

    if not accounts:
        return ("No Payment Users for %s", audit_corp.corporation.corporation_name)

    users = {}

    for user in accounts:
        alts = user.get_alt_ids()
        users[user] = alts

    journal = CorporationWalletJournalEntry.objects.filter(
        division__corporation=audit_corp, ref_type__in=["player_donation"]
    ).order_by("-date")

    _current_entry_ids = set(
        Payments.objects.filter(account__corporation=audit_corp).values_list(
            "entry_id", flat=True
        )
    )
    with transaction.atomic():
        items = []
        logs_items = []
        for entry in journal:
            # Skip if already processed
            if entry.entry_id in _current_entry_ids:
                continue
            for user, alts in users.items():
                if entry.first_party.id in alts:
                    payment_item = Payments(
                        entry_id=entry.entry_id,
                        name=user.name,
                        account=user,
                        amount=entry.amount,
                        request_status=Payments.RequestStatus.PENDING,
                        date=entry.date,
                        reason=entry.reason,
                    )
                    items.append(payment_item)

        payments = Payments.objects.bulk_create(items, ignore_conflicts=True)

        for payment in payments:
            try:
                payment = Payments.objects.get(
                    entry_id=payment.entry_id, account=payment.account
                )
            except Payments.DoesNotExist:
                continue

            log_items = PaymentHistory(
                user=payment.account.user,
                payment=payment,
                action=PaymentHistory.Actions.STATUS_CHANGE,
                new_status=Payments.RequestStatus.PENDING,
                comment=PaymentHistory.SystemText.ADDED,
            )
            logs_items.append(log_items)

        PaymentHistory.objects.bulk_create(logs_items, ignore_conflicts=True)

        audit_corp.last_update_payments = timezone.now()
        audit_corp.save()

    logger.debug(
        "Finished %s Payments for %s",
        len(items),
        audit_corp.corporation.corporation_name,
    )

    return ("Finished Payments for %s", audit_corp.corporation.corporation_name)


def update_corporation_payments_filter(corp_id, runs=0):
    """Update Filtered Payments for Corporation"""
    audit_corp = OwnerAudit.objects.get(corporation__corporation_id=corp_id)

    logger.debug(
        "Updating payment system for: %s",
        audit_corp.corporation.corporation_name,
    )

    payments = Payments.objects.filter(
        account__corporation=audit_corp, request_status=Payments.RequestStatus.PENDING
    )

    _current_payment_ids = set(payments.values_list("id", flat=True))
    _automatic_payment_ids = []

    # Check for any automatic payments
    try:
        filters = SmartGroup.objects.get(corporation=audit_corp)
        if filters:
            payments = filters.filter(payments)
            for payment in payments:
                if payment.request_status == Payments.RequestStatus.PENDING:
                    # Ensure all transfers are processed in a single transaction
                    with transaction.atomic():
                        payment.request_status = Payments.RequestStatus.APPROVED
                        payment.reviser = "System"

                        # Update payment pool for user
                        PaymentSystem.objects.filter(
                            corporation=audit_corp, user=payment.account.user
                        ).update(deposit=payment.account.deposit + payment.amount)

                        payment.save()

                        PaymentHistory(
                            user=payment.account.user,
                            payment=payment,
                            action=PaymentHistory.Actions.STATUS_CHANGE,
                            new_status=Payments.RequestStatus.APPROVED,
                            comment=PaymentHistory.SystemText.AUTOMATIC,
                        ).save()

                        runs = runs + 1
                        _automatic_payment_ids.append(payment.pk)
    except SmartGroup.DoesNotExist:
        pass

    # Check for any payments that need approval
    needs_approval = _current_payment_ids - set(_automatic_payment_ids)
    approvals = Payments.objects.filter(
        id__in=needs_approval, request_status=Payments.RequestStatus.PENDING
    )

    for payment in approvals:
        payment.request_status = Payments.RequestStatus.NEEDS_APPROVAL
        payment.save()

        PaymentHistory(
            user=payment.account.user,
            payment=payment,
            action=PaymentHistory.Actions.STATUS_CHANGE,
            new_status=Payments.RequestStatus.NEEDS_APPROVAL,
            comment=PaymentHistory.SystemText.REVISER,
        ).save()

        runs = runs + 1

    audit_corp.last_update_payment_system = timezone.now()
    audit_corp.save()

    logger.debug(
        "Finished %s: Payment System entrys for %s",
        runs,
        audit_corp.corporation.corporation_name,
    )

    return ("Finished Payment System for %s", audit_corp.corporation.corporation_name)


def update_payday(corp_id, runs=0):
    """Update Payday for Corporation"""
    audit_corp = OwnerAudit.objects.get(corporation__corporation_id=corp_id)

    logger.debug(
        "Updating payday for: %s",
        audit_corp.corporation.corporation_name,
    )

    payday = PaymentSystem.objects.filter(
        corporation=audit_corp, status=PaymentSystem.Status.ACTIVE
    )

    for user in payday:
        if user.last_paid is None:
            # First Period is free
            user.last_paid = timezone.now()
        if timezone.now() - user.last_paid >= timezone.timedelta(
            days=audit_corp.tax_period
        ):
            user.deposit -= audit_corp.tax_amount
            user.last_paid = timezone.now()
            runs = runs + 1
        user.save()

    audit_corp.last_update_payday = timezone.now()
    audit_corp.save()

    logger.debug(
        "Finished %s: Payday for %s",
        runs,
        audit_corp.corporation.corporation_name,
    )

    return ("Finished Payday for %s", audit_corp.corporation.corporation_name)



"""App Tasks"""

import datetime
import logging

from celery import chain, shared_task

from django.utils import timezone

from allianceauth.services.tasks import QueueOnce

from taxsystem import app_settings
from taxsystem.decorators import when_esi_is_available
from taxsystem.models.tax import OwnerAudit
from taxsystem.task_helpers.payment_helpers import (
    update_corporation_payments,
    update_corporation_payments_filter,
    update_payday,
)
from taxsystem.task_helpers.tax_helpers import update_corporation_members
from taxsystem.task_helpers.wallet_helpers import update_corporation_wallet_division

logger = logging.getLogger(__name__)

MAX_RETRIES_DEFAULT = 3

# Default params for all tasks.
TASK_DEFAULTS = {
    "time_limit": app_settings.TAXSYSTEM_TASKS_TIME_LIMIT,
    "max_retries": MAX_RETRIES_DEFAULT,
}

# Default params for tasks that need run once only.
TASK_DEFAULTS_ONCE = {**TASK_DEFAULTS, **{"base": QueueOnce}}

_update_taxsystem_params = {
    **TASK_DEFAULTS_ONCE,
    **{"once": {"keys": ["corp_id"], "graceful": True}},
}


@shared_task
@when_esi_is_available
def update_all_taxsytem(runs: int = 0):
    corps = OwnerAudit.objects.select_related("corporation").all()
    for corp in corps:
        update_corp.apply_async(args=[corp.corporation.corporation_id])
        runs = runs + 1


@shared_task(bind=True, base=QueueOnce)
def update_corp(self, corp_id, force_refresh=False):  # pylint: disable=unused-argument
    class SkipDates:
        """Skip Dates for Updates"""

        wallet = timezone.now() - datetime.timedelta(
            hours=app_settings.TAXSYSTEM_CORP_WALLET_SKIP_DATE
        )
        members = timezone.now() - datetime.timedelta(
            hours=app_settings.TAXSYSTEM_CORP_MEMBERS_SKIP_DATE
        )
        payments = timezone.now() - datetime.timedelta(
            hours=app_settings.TAXSYSTEM_CORP_PAYMENTS_SKIP_DATE
        )
        payment_system = timezone.now() - datetime.timedelta(
            hours=app_settings.TAXSYSTEM_CORP_PAYMENT_SYSTEM_SKIP_DATE
        )
        payment_payday = timezone.now() - datetime.timedelta(
            hours=app_settings.TAXSYSTEM_CORP_PAYMENT_PAYDAY_SKIP_DATE
        )

    corp = OwnerAudit.objects.get(corporation__corporation_id=corp_id)
    logger.debug("Processing Audit Updates for %s", corp.corporation.corporation_name)

    que = []
    mindt = timezone.now() - datetime.timedelta(days=app_settings.TAXSYSTEM_STALE_TIME)
    priority = 7

    if (corp.last_update_wallet or mindt) <= SkipDates.wallet or force_refresh:
        que.append(
            update_corp_wallet.si(corp_id, force_refresh=force_refresh).set(
                priority=priority
            )
        )
    if (corp.last_update_members or mindt) <= SkipDates.members or force_refresh:
        que.append(
            update_corp_members.si(corp_id, force_refresh=force_refresh).set(
                priority=priority
            )
        )
    if (corp.last_update_payments or mindt) <= SkipDates.payments or force_refresh:
        que.append(update_corp_payments.si(corp_id).set(priority=priority))
    if (
        corp.last_update_payment_system or mindt
    ) <= SkipDates.payment_system or force_refresh:
        que.append(update_corp_payments_filter.si(corp_id).set(priority=priority))
    if (corp.last_update_payday or mindt) <= SkipDates.payment_payday or force_refresh:
        que.append(update_corp_payday.si(corp_id).set(priority=priority))

    chain(que).apply_async()
    logger.info("Queued Updates for %s", corp.corporation.corporation_name)


@shared_task(**_update_taxsystem_params)
def update_corp_wallet(
    corp_id,
    force_refresh=False,
):
    return update_corporation_wallet_division(corp_id, force_refresh=force_refresh)


@shared_task(**_update_taxsystem_params)
def update_corp_members(
    corp_id,
    force_refresh=False,
):
    return update_corporation_members(corp_id, force_refresh=force_refresh)


@shared_task(**_update_taxsystem_params)
def update_corp_payments(
    corp_id,
):
    return update_corporation_payments(corp_id)


@shared_task(**_update_taxsystem_params)
def update_corp_payments_filter(
    corp_id,
):
    return update_corporation_payments_filter(corp_id)


@shared_task(**_update_taxsystem_params)
def update_corp_payday(
    corp_id,
):
    return update_payday(corp_id)



from django.contrib.humanize.templatetags.humanize import intcomma
from django.urls import reverse
from django.utils.html import format_html
from django.utils.translation import gettext as _

from taxsystem.api.helpers import generate_button, generate_settings
from taxsystem.models.tax import PaymentSystem


def _payments_info(corporation_id, user: PaymentSystem, perms, request):
    if not perms:
        return ""

    view_payments = generate_button(
        corporation_id,
        "taxsystem/partials/form/button.html",
        user,
        {
            "title": _("View Payments"),
            "icon": "fas fa-info",
            "color": "primary",
            "text": _("View Payments"),
            "modal": "modalViewPaymentsContainer",
            "action": f"/taxsystem/api/corporation/{corporation_id}/character/{user.user.profile.main_character.character_id}/view/payments/",
            "ajax": "ajax_payments",
        },
        request,
    )

    return format_html(f"{intcomma(user.deposit, use_l10n=True)} ISK {view_payments}")


def _payment_system_actions(corporation_id, user: PaymentSystem, perms, request):
    # Check if user has permission to view the actions
    if not perms:
        return ""

    template = "taxsystem/partials/form/button.html"
    url = reverse(
        viewname="taxsystem:switch_user",
        kwargs={"corporation_id": corporation_id, "user_pk": user.pk},
    )

    if user.is_active:
        confirm_text = (
            _("Are you sure to Confirm")
            + f"?<br><span class='fw-bold'>{user.name} "
            + _("Deactivate")
            + "</span>"
        )
        title = _("Deactivate User")
        icon = "fas fa-eye-low-vision"
        color = "warning"
    else:
        confirm_text = (
            _("Are you sure to Confirm")
            + f"?<br><span class='fw-bold'>{user.name} "
            + _("Activate")
            + "</span>"
        )
        title = _("Activate User")
        icon = "fas fa-eye"
        color = "success"

    settings = generate_settings(
        title=title,
        icon=icon,
        color=color,
        text=confirm_text,
        modal="paymentsystem-switchuser",
        action=url,
        ajax="action",
    )
    # Generate the buttons
    actions = []
    actions.append(generate_button(corporation_id, template, user, settings, request))

    actions_html = format_html("".join(actions))
    return format_html('<div class="d-flex justify-content-end">{}</div>', actions_html)


def _get_has_paid_icon(user: PaymentSystem) -> dict:
    if user.is_active:
        has_paid_filter = _("Yes") if user.has_paid else _("No")
        if user.has_paid:
            button = format_html(
                '<button class="btn btn-success btn-sm d-flex align-items-center justify-content-center" '
                'style="height: 30px; width: 30px;" title="{}" data-tooltip-toggle="taxsystem-tooltip">'
                '<i class="fas fa-check"></i></button>',
                _("Paid"),
            )
        else:
            button = format_html(
                '<button class="btn btn-danger btn-sm d-flex align-items-center justify-content-center" '
                'style="height: 30px; width: 30px;" title="{}" data-tooltip-toggle="taxsystem-tooltip">'
                '<i class="fas fa-times"></i></button>',
                _("Unpaid"),
            )
        has_paid = {
            "display": button,
            "sort": has_paid_filter,
            "raw": user.has_paid,
        }
    else:
        button = format_html(
            '<button class="btn btn-warning btn-sm d-flex align-items-center justify-content-center" '
            'style="height: 30px; width: 30px;" title="{}" data-tooltip-toggle="taxsystem-tooltip">'
            '<i class="fas fa-user-slash"></i></button>',
            _("Inactive"),
        )
        has_paid_filter = ""
        has_paid = {
            "display": button,
            "sort": "",
            "raw": "",
        }
    return has_paid



"""PvE Views"""

import logging

# Django
from django.contrib import messages
from django.contrib.auth.decorators import login_required, permission_required
from django.contrib.humanize.templatetags.humanize import intcomma
from django.core.exceptions import ValidationError
from django.core.handlers.wsgi import WSGIRequest
from django.db import IntegrityError, transaction
from django.http import JsonResponse
from django.shortcuts import get_object_or_404, redirect, render
from django.utils.translation import gettext_lazy as _
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_POST
from esi.decorators import token_required

from allianceauth.authentication.decorators import permissions_required
from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo

from taxsystem import forms
from taxsystem.api.helpers import get_corporation, get_manage_permission
from taxsystem.helpers.views import add_info_to_context
from taxsystem.models.logs import AdminLogs, PaymentHistory
from taxsystem.models.tax import Members, OwnerAudit, Payments, PaymentSystem
from taxsystem.tasks import update_corp

logger = logging.getLogger(__name__)


@login_required
@permission_required("taxsystem.basic_access")
def index(request):
    """Index View"""
    return redirect(
        "taxsystem:payments", request.user.profile.main_character.corporation_id
    )


@login_required
@permissions_required(["taxsystem.manage_own_corp", "taxsystem.manage_corps"])
def administration(request, corporation_id):
    """Manage View"""
    if corporation_id is None:
        corporation_id = request.user.profile.main_character.corporation_id

    context = {
        "corporation_id": corporation_id,
        "title": _("Administration"),
        "forms": {
            "accept_request": forms.TaxAcceptForm(),
            "reject_request": forms.TaxRejectForm(),
            "undo_request": forms.TaxUndoForm(),
            "switchuser_request": forms.TaxSwitchUserForm(),
            "delete_request": forms.TaxDeleteForm(),
        },
    }
    context = add_info_to_context(request, context)

    return render(request, "taxsystem/manage.html", context=context)


@login_required
@permission_required("taxsystem.basic_access")
def payments(request, corporation_id):
    """Payments View"""
    if corporation_id is None:
        corporation_id = request.user.profile.main_character.corporation_id

    perms = get_corporation(request, corporation_id)

    if perms is None:
        messages.error(request, _("No corporation found."))

    corporations = OwnerAudit.objects.visible_to(request.user)

    context = {
        "corporation_id": corporation_id,
        "title": _("Payments"),
        "forms": {
            "accept_request": forms.TaxAcceptForm(),
            "reject_request": forms.TaxRejectForm(),
            "undo_request": forms.TaxUndoForm(),
        },
        "corporations": corporations,
    }
    context = add_info_to_context(request, context)

    return render(request, "taxsystem/payments.html", context=context)


@login_required
@permission_required("taxsystem.basic_access")
def own_payments(request, corporation_id=None):
    """Own Payments View"""
    if corporation_id is None:
        corporation_id = request.user.profile.main_character.corporation_id

    perms = get_corporation(request, corporation_id)

    if perms is None:
        messages.error(request, _("No corporation found."))

    corporations = OwnerAudit.objects.visible_to(request.user)

    context = {
        "corporation_id": corporation_id,
        "title": _("Own Payments"),
        "corporations": corporations,
    }
    context = add_info_to_context(request, context)

    return render(request, "taxsystem/own-payments.html", context=context)


@login_required
@permission_required("taxsystem.create_access")
@token_required(scopes=OwnerAudit.get_esi_scopes())
def add_corp(request, token):
    char = get_object_or_404(EveCharacter, character_id=token.character_id)
    corp, _ = EveCorporationInfo.objects.get_or_create(
        corporation_id=char.corporation_id,
        defaults={
            "member_count": 0,
            "corporation_ticker": char.corporation_ticker,
            "corporation_name": char.corporation_name,
        },
    )

    owner, created = OwnerAudit.objects.update_or_create(
        corporation=corp,
        defaults={
            "name": corp.corporation_name,
            "active": True,
        },
    )

    if created:
        AdminLogs(
            user=request.user,
            corporation=owner,
            action=AdminLogs.Actions.ADD,
            comment=_("Added to Tax System"),
        ).save()

    update_corp.apply_async(
        args=[char.corporation_id], kwargs={"force_refresh": True}, priority=6
    )
    msg = _("{corporation_name} successfully added/updated to Tax System").format(
        corporation_name=corp.corporation_name,
    )
    messages.info(request, msg)
    return redirect("taxsystem:index")


@login_required
@permissions_required(["taxsystem.manage_own_corp", "taxsystem.manage_corps"])
@require_POST
def approve_payment(request: WSGIRequest, corporation_id: int, payment_pk: int):
    msg = _("Invalid Method")
    corp = get_corporation(request, corporation_id)

    perms = get_manage_permission(request, corporation_id)
    if not perms:
        msg = _("Permission Denied")
        return JsonResponse(
            data={"success": False, "message": msg}, status=403, safe=False
        )

    try:
        with transaction.atomic():
            form = forms.TaxAcceptForm(data=request.POST)
            if form.is_valid():
                reason = form.cleaned_data["accept_info"]
                payment = Payments.objects.get(account__corporation=corp, pk=payment_pk)
                if payment.is_pending or payment.is_needs_approval:
                    payment.request_status = Payments.RequestStatus.APPROVED
                    payment.reviser = request.user.profile.main_character.character_name
                    payment.save()

                    account = PaymentSystem.objects.get(
                        corporation=corp, user=payment.account.user
                    )
                    account.deposit += payment.amount
                    account.save()
                    PaymentHistory(
                        user=request.user,
                        payment=payment,
                        action=PaymentHistory.Actions.STATUS_CHANGE,
                        comment=reason,
                        new_status=Payments.RequestStatus.APPROVED,
                    ).save()
                    return JsonResponse(
                        data={"success": True, "message": msg}, status=200, safe=False
                    )
    except IntegrityError:
        msg = _("Transaction failed. Please try again.")
    return JsonResponse(data={"success": False, "message": msg}, status=400, safe=False)


@login_required
@permissions_required(["taxsystem.manage_own_corp", "taxsystem.manage_corps"])
@require_POST
def undo_payment(request: WSGIRequest, corporation_id: int, payment_pk: int):
    msg = _("Invalid Method")
    corp = get_corporation(request, corporation_id)

    perms = get_manage_permission(request, corporation_id)
    if not perms:
        msg = _("Permission Denied")
        return JsonResponse(
            data={"success": False, "message": msg}, status=403, safe=False
        )

    try:
        with transaction.atomic():
            form = forms.TaxUndoForm(data=request.POST)
            if form.is_valid():
                reason = form.cleaned_data["undo_reason"]
                payment = Payments.objects.get(account__corporation=corp, pk=payment_pk)
                if payment.is_approved or payment.is_rejected:
                    # Ensure that the payment is not rejected
                    if not payment.is_rejected:
                        account = PaymentSystem.objects.get(
                            corporation=corp, user=payment.account.user
                        )
                        account.deposit -= payment.amount
                        account.save()
                    payment.request_status = Payments.RequestStatus.PENDING
                    payment.reviser = ""
                    payment.save()
                    PaymentHistory(
                        user=request.user,
                        payment=payment,
                        action=PaymentHistory.Actions.STATUS_CHANGE,
                        comment=reason,
                        new_status=Payments.RequestStatus.PENDING,
                    ).save()
                    return JsonResponse(
                        data={"success": True, "message": msg}, status=200, safe=False
                    )
    except IntegrityError:
        msg = _("Transaction failed. Please try again.")
    return JsonResponse(data={"success": False, "message": msg}, status=400, safe=False)


@login_required
@permissions_required(["taxsystem.manage_own_corp", "taxsystem.manage_corps"])
@require_POST
def reject_payment(request: WSGIRequest, corporation_id: int, payment_pk: int):
    msg = _("Invalid Method")
    corp = get_corporation(request, corporation_id)

    perms = get_manage_permission(request, corporation_id)
    if not perms:
        msg = _("Permission Denied")
        return JsonResponse(
            data={"success": False, "message": msg}, status=403, safe=False
        )

    try:
        with transaction.atomic():
            form = forms.TaxRejectForm(data=request.POST)
            if form.is_valid():
                reason = form.cleaned_data["reject_reason"]
                payment = Payments.objects.get(account__corporation=corp, pk=payment_pk)
                if payment.is_pending or payment.is_needs_approval:
                    payment.request_status = Payments.RequestStatus.REJECTED
                    payment.reviser = request.user.profile.main_character.character_name
                    payment.save()

                    account = PaymentSystem.objects.get(
                        corporation=corp, user=payment.account.user
                    )
                    account.save()
                    msg = _("Payment ID: %s - Amount %s - Name: %s rejected") % (
                        payment.pk,
                        intcomma(payment.amount),
                        payment.name,
                    )

                    PaymentHistory(
                        user=request.user,
                        payment=payment,
                        action=PaymentHistory.Actions.STATUS_CHANGE,
                        comment=reason,
                        new_status=Payments.RequestStatus.REJECTED,
                    ).save()
                    return JsonResponse(
                        data={"success": True, "message": msg}, status=200, safe=False
                    )
    except IntegrityError:
        msg = _("Transaction failed. Please try again.")
    return JsonResponse(data={"success": False, "message": msg}, status=400, safe=False)


@login_required
@permissions_required(["taxsystem.manage_own_corp", "taxsystem.manage_corps"])
@require_POST
def switch_user(request: WSGIRequest, corporation_id: int, user_pk: int):
    msg = _("Invalid Method")
    corp = get_corporation(request, corporation_id)

    perms = get_manage_permission(request, corporation_id)
    if not perms:
        msg = _("Permission Denied")
        return JsonResponse(
            data={"success": False, "message": msg}, status=403, safe=False
        )

    try:
        with transaction.atomic():
            form = forms.TaxSwitchUserForm(data=request.POST)
            if form.is_valid():
                payment_system = PaymentSystem.objects.get(corporation=corp, pk=user_pk)
                if payment_system.is_active:
                    payment_system.status = PaymentSystem.Status.DEACTIVATED
                    msg = _("Payment System User: %s deactivated") % payment_system.name
                else:
                    payment_system.status = PaymentSystem.Status.ACTIVE
                    msg = _("Payment System User: %s activated") % payment_system.name

                AdminLogs(
                    user=request.user,
                    corporation=corp,
                    action=AdminLogs.Actions.CHANGE,
                    comment=msg,
                ).save()
                payment_system.save()
            return JsonResponse(
                data={"success": True, "message": msg}, status=200, safe=False
            )
    except IntegrityError:
        msg = _("Transaction failed. Please try again.")
    return JsonResponse(data={"success": False, "message": msg}, status=400, safe=False)


@csrf_exempt
def update_tax_amount(request: WSGIRequest, corporation_id: int):
    if request.method == "POST":
        value = float(request.POST.get("value"))
        msg = _("Please enter a valid number")
        try:
            if value < 0:
                return JsonResponse({"message": msg}, status=400)
        except ValueError:
            return JsonResponse({"message": msg}, status=400)

        corp = get_corporation(request, corporation_id)

        perms = get_manage_permission(request, corporation_id)

        if not perms:
            return JsonResponse({"message": _("Permission Denied")}, status=403)

        try:
            corp.tax_amount = value
            corp.save()
            msg = _(f"Tax Amount from {corp.name} updated to {value}")
            AdminLogs(
                user=request.user,
                corporation=corp,
                action=AdminLogs.Actions.CHANGE,
                comment=msg,
            ).save()
        except ValidationError:
            return JsonResponse({"message": msg}, status=400)
        return JsonResponse({"message": msg}, status=200)
    return JsonResponse({"message": _("Invalid request method")}, status=405)


@csrf_exempt
def update_tax_period(request: WSGIRequest, corporation_id: int):
    if request.method == "POST":
        value = float(request.POST.get("value"))
        msg = _("Please enter a valid number")
        try:
            if value < 0:
                return JsonResponse({"message": msg}, status=400)
        except ValueError:
            return JsonResponse({"message": msg}, status=400)

        corp = get_corporation(request, corporation_id)

        perms = get_manage_permission(request, corporation_id)

        if not perms:
            return JsonResponse({"message": _("Permission Denied")}, status=403)

        try:
            corp.tax_period = value
            corp.save()
            msg = _(f"Tax Period from {corp.name} updated to {value}")
            AdminLogs(
                user=request.user,
                corporation=corp,
                action=AdminLogs.Actions.CHANGE,
                comment=msg,
            ).save()
        except ValidationError:
            return JsonResponse({"message": msg}, status=400)
        return JsonResponse({"message": msg}, status=200)
    return JsonResponse({"message": _("Invalid request method")}, status=405)


@login_required
@permissions_required(["taxsystem.manage_own_corp", "taxsystem.manage_corps"])
@require_POST
def delete_user(request: WSGIRequest, corporation_id: int, member_pk: int):
    msg = _("Invalid Method")
    corp = get_corporation(request, corporation_id)

    perms = get_manage_permission(request, corporation_id)
    if not perms:
        msg = _("Permission Denied")
        return JsonResponse(
            data={"success": False, "message": msg}, status=403, safe=False
        )

    form = forms.TaxDeleteForm(data=request.POST)
    if form.is_valid():
        reason = form.cleaned_data["delete_reason"]
        member = Members.objects.get(corporation=corp, pk=member_pk)
        if member.is_missing:
            msg = _(f"Member {member.character_name} deleted")
            member.delete()
            AdminLogs(
                user=request.user,
                corporation=corp,
                action=AdminLogs.Actions.DELETE,
                comment=reason,
            ).save()
            return JsonResponse(
                data={"success": True, "message": msg}, status=200, safe=False
            )
    return JsonResponse(data={"success": False, "message": msg}, status=400, safe=False)



"""Tax Helpers"""

import logging

from django.utils import timezone
from eveuniverse.models import EveEntity

from allianceauth.authentication.models import UserProfile

from taxsystem.models.tax import (
    Members,
    OwnerAudit,
    PaymentSystem,
)
from taxsystem.providers import esi
from taxsystem.task_helpers.etag_helpers import (
    HTTPGatewayTimeoutError,
    NotModifiedError,
    etag_results,
)
from taxsystem.task_helpers.general_helpers import get_corp_token

logger = logging.getLogger(__name__)


# pylint: disable=too-many-locals
def update_corporation_members(corp_id, force_refresh=False):
    """Update corporation members"""
    audit_corp = OwnerAudit.objects.get(corporation__corporation_id=corp_id)

    logger.debug(
        "Updating members for: %s",
        audit_corp.corporation.corporation_name,
    )

    req_scopes = [
        "esi-corporations.read_corporation_membership.v1",
        "esi-corporations.track_members.v1",
    ]

    req_roles = ["CEO", "Director"]

    token = get_corp_token(corp_id, req_scopes, req_roles)

    # pylint: disable=duplicate-code
    if not token:
        logger.debug("No valid token for: %s", audit_corp.corporation.corporation_name)
        return "No Tokens"

    # Check Payment Accounts

    check_payment_accounts(corp_id)

    try:
        _current_members_ids = set(
            Members.objects.filter(corporation=audit_corp).values_list(
                "character_id", flat=True
            )
        )
        members_ob = (
            esi.client.Corporation.get_corporations_corporation_id_membertracking(
                corporation_id=audit_corp.corporation.corporation_id,
            )
        )

        members = etag_results(members_ob, token, force_refresh=force_refresh)

        _esi_members_ids = [member.get("character_id") for member in members]
        _old_members = []
        _new_members = []

        characters = EveEntity.objects.bulk_resolve_names(_esi_members_ids)

        for member in members:
            character_id = member.get("character_id")
            joined = member.get("start_date")
            logon_date = member.get("logon_date")
            logged_off = member.get("logoff_date")
            character_name = characters.to_name(character_id)
            member_item = Members(
                corporation=audit_corp,
                character_id=character_id,
                character_name=character_name,
                joined=joined,
                logon=logon_date,
                logged_off=logged_off,
                status=Members.States.ACTIVE,
            )
            if character_id in _current_members_ids:
                _old_members.append(member_item)
            else:
                _new_members.append(member_item)

        # Set missing members
        old_member_ids = {member.character_id for member in _old_members}
        missing_members_ids = _current_members_ids - old_member_ids

        if missing_members_ids:
            Members.objects.filter(
                corporation=audit_corp, character_id__in=missing_members_ids
            ).update(status=Members.States.MISSING)
            logger.debug(
                "Marked %s missing members for: %s",
                len(missing_members_ids),
                audit_corp.corporation.corporation_name,
            )
        if _old_members:
            Members.objects.bulk_update(
                _old_members,
                ["character_name", "status", "logon", "logged_off"],
            )
            logger.debug(
                "Updated %s members for: %s",
                len(_old_members),
                audit_corp.corporation.corporation_name,
            )
        if _new_members:
            Members.objects.bulk_create(_new_members, ignore_conflicts=True)
            logger.debug(
                "Added %s new members for: %s",
                len(_new_members),
                audit_corp.corporation.corporation_name,
            )

        # Update payment accounts
        update_payment_accounts(corp_id, _esi_members_ids)

        logger.info(
            "Corp %s - Old Members: %s, New Members: %s, Missing: %s",
            audit_corp.name,
            len(_old_members),
            len(_new_members),
            len(missing_members_ids),
        )
    except NotModifiedError:
        logger.debug(
            "No changes detected for: %s", audit_corp.corporation.corporation_name
        )
    except HTTPGatewayTimeoutError:
        logger.debug(
            "ESI Timeout skipping Members for: %s",
            audit_corp.corporation.corporation_name,
        )
        return (
            "ESI Timeout skipping Members for %s",
            audit_corp.corporation.corporation_name,
        )
    audit_corp.last_update_members = timezone.now()
    audit_corp.save()

    return ("Finished Members for %s", audit_corp.corporation.corporation_name)


def update_payment_accounts(corp_id: int, members_ids: list[int]):
    """Update payment accounts for a corporation."""
    audit_corp = OwnerAudit.objects.get(corporation__corporation_id=corp_id)

    logger.debug(
        "Updating Payment Accounts for: %s",
        audit_corp.corporation.corporation_name,
    )

    accounts = UserProfile.objects.filter(
        main_character__isnull=False,
        main_character__corporation_id=audit_corp.corporation.corporation_id,
    ).select_related(
        "user__profile__main_character",
        "main_character__character_ownership",
        "main_character__character_ownership__user__profile",
        "main_character__character_ownership__user__profile__main_character",
    )

    members = Members.objects.filter(corporation=audit_corp)

    if not accounts:
        logger.debug(
            "No valid accounts for: %s", audit_corp.corporation.corporation_name
        )
        return "No Accounts"

    items = []

    for account in accounts:
        alts = set(
            account.user.character_ownerships.all().values_list(
                "character__character_id", flat=True
            )
        )
        main = account.main_character

        relevant_alts = alts.intersection(members_ids)
        for alt in relevant_alts:
            members_ids.remove(alt)
            if alt != main.character_id:
                # Update the status of the member to alt
                members.filter(character_id=alt).update(status=Members.States.IS_ALT)
        try:
            existing_payment_system = PaymentSystem.objects.get(
                user=account.user, corporation=audit_corp
            )

            if existing_payment_system.status != PaymentSystem.Status.DEACTIVATED:
                existing_payment_system.status = PaymentSystem.Status.ACTIVE
                existing_payment_system.save()
        except PaymentSystem.DoesNotExist:
            items.append(
                PaymentSystem(
                    name=main.character_name,
                    corporation=audit_corp,
                    user=account.user,
                    status=PaymentSystem.Status.ACTIVE,
                )
            )

    if members_ids:
        # Mark members without accounts
        for member_id in members_ids:
            members.filter(character_id=member_id).update(
                status=Members.States.NOACCOUNT
            )

        logger.debug(
            "Marked %s members without accounts for: %s",
            len(members_ids),
            audit_corp.corporation.corporation_name,
        )

    if items:
        PaymentSystem.objects.bulk_create(items, ignore_conflicts=True)
        logger.info(
            "Added %s new payment users for: %s",
            len(items),
            audit_corp.corporation.corporation_name,
        )
    else:
        logger.debug(
            "No new payment user for: %s",
            audit_corp.corporation.corporation_name,
        )

    return ("Finished payment system for %s", audit_corp.corporation.corporation_name)


def check_payment_accounts(corp_id: int):
    """Check payment accounts for a corporation."""
    audit_corp = OwnerAudit.objects.get(corporation__corporation_id=corp_id)

    logger.debug(
        "Checking Payment Accounts for: %s",
        audit_corp.corporation.corporation_name,
    )

    accounts = UserProfile.objects.filter(
        main_character__isnull=False,
    ).select_related(
        "user__profile__main_character",
        "main_character__character_ownership",
        "main_character__character_ownership__user__profile",
        "main_character__character_ownership__user__profile__main_character",
    )

    if not accounts:
        logger.debug(
            "No valid accounts for skipping Check: %s",
            audit_corp.corporation.corporation_name,
        )
        return "No Accounts"

    for account in accounts:
        main_corporation_id = account.main_character.corporation_id

        try:
            payment_system = PaymentSystem.objects.get(
                user=account.user, corporation=audit_corp
            )
            payment_system_corp_id = (
                payment_system.corporation.corporation.corporation_id
            )
            # Check if the user is no longer in the same corporation
            if (
                not payment_system.is_missing
                and not payment_system_corp_id == main_corporation_id
            ):
                payment_system.status = PaymentSystem.Status.MISSING
                payment_system.save()
                logger.info(
                    "User %s is no longer in Corp marked as Missing",
                    payment_system.name,
                )
            # Check if the user changed to a existing corporation Payment System
            elif (
                payment_system.is_missing
                and payment_system_corp_id != main_corporation_id
            ):
                try:
                    new_audit_corp = OwnerAudit.objects.get(
                        corporation__corporation_id=main_corporation_id
                    )
                    payment_system.corporation = new_audit_corp
                    payment_system.deposit = 0
                    payment_system.status = PaymentSystem.Status.ACTIVE
                    payment_system.last_paid = None
                    payment_system.save()
                    logger.info(
                        "User %s is now in Corp %s",
                        payment_system.name,
                        new_audit_corp.corporation.corporation_name,
                    )
                except OwnerAudit.DoesNotExist:
                    continue
            elif (
                payment_system.is_missing
                and payment_system_corp_id == main_corporation_id
            ):
                payment_system.status = PaymentSystem.Status.ACTIVE
                payment_system.notice = None
                payment_system.deposit = 0
                payment_system.last_paid = None
                payment_system.save()
                logger.info(
                    "User %s is back in Corp %s",
                    payment_system.name,
                    payment_system.corporation.corporation.corporation_name,
                )
        except PaymentSystem.DoesNotExist:
            logger.debug(
                "No Payment System for %s - %s",
                account.user.username,
                audit_corp.corporation.corporation_name,
            )
            continue
    return (
        "Finished checking Payment Accounts for %s",
        audit_corp.corporation.corporation_name,
    )



# Django
import logging

from django.db import models

logger = logging.getLogger(__name__)


class WalletQuerySet(models.QuerySet):
    pass


class WalletManagerBase(models.Manager):
    pass


WalletManager = WalletManagerBase.from_queryset(WalletQuerySet)
